# -*- encoding: utf-8 -*-
"""
KERI
keri.app.delegating module

module for enveloping and forwarding KERI message
"""
import json
import logging
import random

from hio import help
from hio.base import doing
from hio.help import decking

from . import keeping, habbing, agenting, obtaining
from .forwarding import forward
from .habbing import Habitat
from .. import kering
from ..core import eventing, coring, parsing
from ..db import basing

logger = help.ogler.getLogger()


class Delegatey:
    def __init__(self, name, db, ks, msgs=None, posts=None):
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.posts = posts if posts is not None else decking.Deck()
        self.name = name
        self.db = db
        self.ks = ks
        self.hab = None
        self.notified = False

    def processMessage(self, msg):
        salt = coring.Salter(raw=msg["salt"].encode("utf-8")).qb64
        seed = msg["seed"] if "seed" in msg else None
        wits = msg["wits"] if "wits" in msg else None
        toad = msg["toad"] if "toad" in msg else None
        icount = msg["icount"] if "icount" in msg else None
        isith = msg["isith"] if "isith" in msg else None
        ncount = msg["ncount"] if "ncount" in msg else None
        nsith = msg["nsith"] if "nsith" in msg else None
        delpre = msg["delpre"] if "delpre" in msg else None

        if self.hab is None:
            self.hab = Habitat(name=self.name, db=self.db, ks=self.ks, seed=seed, salt=salt, icount=icount, isith=isith,
                               ncount=ncount, nsith=nsith, toad=toad, wits=wits, delpre=delpre, )

        self.posts.append(dict(srdr=self.hab.delserder, sigers=self.hab.delsigers))

    def genQuery(self, pre):
        query = dict()
        query['i'] = pre
        serder = eventing.query(res="logs", qry=query)

        msg = bytearray(serder.raw)  # make copy into new bytearray so can be deleted
        msg.extend(coring.Counter(coring.CtrDex.TransLastIdxSigGroups, count=1).qb64b)
        msg.extend(pre.encode("utf-8"))

        counter = coring.Counter(code=coring.CtrDex.ControllerIdxSigs,
                                 count=len(self.hab.delsigers))
        msg.extend(counter.qb64b)
        for siger in self.hab.delsigers:
            msg.extend(siger.qb64b)
        return msg


class InceptDoer(doing.DoDoer):
    def __init__(self, name, msgs=None, cues=None, ):
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.ks = keeping.Keeper(name=name, temp=False)  # not opened by default, doer opens
        self.ksDoer = keeping.KeeperDoer(keeper=self.ks)  # doer do reopens if not opened and closes
        self.db = basing.Baser(name=name, temp=False, reload=True)  # not opened by default, doer opens
        self.dbDoer = basing.BaserDoer(baser=self.db)  # doer do reopens if not opened and closes

        doers = [
            self.ksDoer,
            self.dbDoer,
            doing.doify(self.msgDo),
            doing.doify(self.postDo),
        ]
        self.delegatey = Delegatey(name=name, db=self.db, ks=self.ks, msgs=self.msgs)
        super(InceptDoer, self).__init__(doers=doers)

    def escrowDo(self, tymth, tock=0.0):
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            self.delegatey.hab.kvy.processEscrows()
            yield self.tock

    def processKvyCues(self, tymth, tock=0.0):
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.delegatey.hab.kvy.cues:
                cue = self.delegatey.hab.kvy.cues.popleft()
                if cue["kin"] == "delegatage":
                    delpre = cue["delpre"]
                    delwits = obtaining.getwitnessesforprefix(delpre)

                    qry = self.delegatey.genQuery(pre=delpre)
                    witer = agenting.TCPWitnesser(self.delegatey.hab, random.choice(delwits), lax=True, local=False)
                    witer.msgs.append(qry)
                    self.extend([witer])
                elif cue["kin"] == "psUnescrow":
                    self.delegatey.hab.delegationAccepted()
                    evt = self.delegatey.hab.makeOwnEvent(sn=0)
                    witDoer = agenting.WitnessReceiptor(hab=self.delegatey.hab, msg=evt, klas=agenting.TCPWitnesser)
                    self.extend([witDoer])
                    while not witDoer.done:
                        yield self.tock

                    self.remove([witDoer])
                    self.cues.append(dict(delegator=self.delegatey.hab.delpre, pre=self.delegatey.hab.pre))


                yield self.tock
            yield self.tock

    def msgDo(self, tymth, tock=0.0):
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.delegatey.msgs:
                msg = self.delegatey.msgs.popleft()
                try:
                    if "name" not in msg:
                        msg["name"] = self.delegatey.name

                    self.delegatey.processMessage(msg=msg)
                except (kering.MissingAnchorError, Exception) as ex:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception("delegation incept message error: %s\n", ex)
                    else:
                        logger.error("delegation incept message error: %s\n", ex)
                yield self.tock
            yield self.tock


    def postDo(self, tymth, tock=0.0):
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.delegatey.posts:
                post = self.delegatey.posts.popleft()
                delsrdr = post["srdr"]
                fwd = forward(pre=delsrdr.ked["di"], topic="delegate", serder=delsrdr)
                evt = eventing.messagize(serder=fwd, sigers=post["sigers"])

                delwits = obtaining.getwitnessesforprefix(delsrdr.ked["di"])
                wit = random.choice(delwits)

                witer = agenting.HttpWitnesser(hab=self.delegatey.hab, wit=wit)
                witer.msgs.append(bytearray(evt))
                self.extend([witer])

                while len(witer.sent) == 0:
                    yield self.tock

                self.extend([doing.doify(self.processKvyCues), doing.doify(self.escrowDo), ])
                # self.remove([witer])

                yield self.tock
            yield self.tock


class RotateDoer(doing.DoDoer):
    """
    DoDoer instance that launches the environment and dependencies needed to create and disseminate
    the inception event for a delegated identifier.
    """

    def __init__(self, name, sealFile, data, **kwa):
        """
        Creates the DoDoer needed to create the seal for a delegated identifier.

        Parameters
            name (str): Name of the local identifier environment

        """

        ks = keeping.Keeper(name=name, temp=False)  # not opened by default, doer opens
        self.ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
        db = basing.Baser(name=name, temp=False, reload=True)  # not opened by default, doer opens
        self.dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

        hab = habbing.Habitat(name=name, ks=ks, db=db, temp=False, create=False)
        self.habDoer = habbing.HabitatDoer(habitat=hab)

        doers = [self.ksDoer, self.dbDoer, self.habDoer, doing.doify(self.rotateDo, **kwa)]
        self.hab = hab
        self.data = data
        self.sealFile = sealFile
        self.delegatorPrefix = kwa["delegatorPrefix"]

        super(RotateDoer, self).__init__(doers=doers)

    def rotateDo(self, tymth, tock=0.0, **kwa):
        # start enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)  # finish enter context

        delPre = self.data[0]["i"]
        delK = self.hab.kevers[delPre]

        verfers, digers, cst, nst = self.hab.mgr.rotate(pre=delK.prefixer.qb64, temp=False)
        rotSrdr = eventing.deltate(pre=delK.prefixer.qb64,
                                   keys=[verfer.qb64 for verfer in verfers],
                                   dig=delK.serder.diger.qb64,
                                   sn=delK.sn + 1,
                                   nxt=coring.Nexter(digs=[diger.qb64 for diger in digers]).qb64)

        seal = dict(i=rotSrdr.pre,
                    s=rotSrdr.ked["s"],
                    d=rotSrdr.dig)

        with open(self.sealFile, "w") as f:
            f.write(json.dumps(seal, indent=4))

        witq = agenting.WitnessInquisitor(hab=self.hab, klas=agenting.TCPWitnesser)
        self.extend([witq])

        print("Hello, could someone approve my delegated rotation, please?")

        while self.delegatorPrefix not in self.hab.kevers or self.hab.kevers[self.delegatorPrefix].sn < 2:
            witq.query(self.delegatorPrefix)
            yield self.tock

        sigers = self.hab.mgr.sign(ser=rotSrdr.raw, verfers=verfers)
        msg = bytearray(rotSrdr.raw)
        counter = coring.Counter(code=coring.CtrDex.ControllerIdxSigs,
                                 count=len(sigers))
        msg.extend(counter.qb64b)
        for siger in sigers:
            msg.extend(siger.qb64b)
        counter = coring.Counter(code=coring.CtrDex.SealSourceCouples,
                                 count=1)
        msg.extend(counter.qb64b)

        event = self.hab.kevers[rotSrdr.pre]
        seqner = coring.Seqner(sn=event.sn)
        msg.extend(seqner.qb64b)
        msg.extend(event.serder.diger.qb64b)

        delKvy = eventing.Kevery(db=self.hab.db, lax=True)
        parsing.Parser().parseOne(ims=bytearray(msg), kvy=delKvy)

        while rotSrdr.pre not in delKvy.kevers:
            yield self.tock

        print("Successfully rotated delegate identifier keys", rotSrdr.pre)
        print("Public key", rotSrdr.verfers[0].qb64)

        self.remove([self.ksDoer, self.dbDoer, self.habDoer, witq])

        return



# -*- encoding: utf-8 -*-
"""
keri.app.storing module

"""
import random

import falcon
from hio.base import doing
from hio.core import http
from hio.help import helping, Hict, decking

from . import obtaining, forwarding, httping, agenting
from .. import help
from ..core import coring
from ..core.coring import MtrDex
from ..db import dbing, subing

logger = help.ogler.getLogger()



class Mailboxer(dbing.LMDBer):
    """
    Mailboxer stores exn messages in order and provider iterator access at an index.

    """
    TailDirPath = "keri/mbx"
    AltTailDirPath = ".keri/mbx"
    TempPrefix = "keri_mbx_"

    def __init__(self, name="mbx", headDirPath=None, reopen=True, **kwa):
        """

        Parameters:
            headDirPath:
            perm:
            reopen:
            kwa:
        """
        self.tpcs = None
        self.msgs = None

        super(Mailboxer, self).__init__(name=name, headDirPath=headDirPath, reopen=reopen, **kwa)

    def reopen(self, **kwa):
        """

        :param kwa:
        :return:
        """
        super(Mailboxer, self).reopen(**kwa)

        self.tpcs = self.env.open_db(key=b'tpcs.', dupsort=True)
        self.msgs = subing.Suber(db=self, subkey='msgs.')  # key states

        return self.env

    def delTopic(self, key):
        """
        Use snKey()
        Deletes value at key.
        Returns True If key exists in database Else False
        """
        return self.delIoSetVals(self.tpcs, key)

    def appendToTopic(self, topic, val):
        """
        Return first seen order number int, fn, of appended entry.
        Computes fn as next fn after last entry.

        Append val to end of db entries with same topic but with fn incremented by
        1 relative to last preexisting entry at pre.

        Parameters:
            topic is bytes identifier prefix/topic for message
            val is event digest
        """
        return self.appendIoSetVal(db=self.tpcs, key=topic, val=val)

    def getTopicMsgs(self, topic, fn=0):
        """
        Returns:
             ioset (oset): the insertion ordered set of values at same apparent
             effective key.
             Uses hidden ordinal key suffix for insertion ordering.
             The suffix is appended and stripped transparently.

         Parameters:
             topic (bytes): Apparent effective key
             fn (int) starting index
        """
        if hasattr(topic, "encode"):
            topic = topic.encode("utf-8")

        digs = self.getIoSetVals(db=self.tpcs, key=topic, ion=fn)
        msgs = []
        for dig in digs:
            if msg := self.msgs.get(keys=dig):
                msgs.append(msg.encode("utf-8"))
        return msgs

    def storeMsg(self, topic, msg):
        """
        Add exn event to mailbox of dest identifier

        Parameters:
            msg (bytes):
            topic (qb64b):

        """
        if hasattr(topic, "encode"):
            topic = topic.encode("utf-8")

        if hasattr(msg, "encode"):
            msg = msg.encode("utf-8")


        digb = coring.Diger(ser=msg, code=MtrDex.Blake3_256).qb64b
        self.appendToTopic(topic=topic, val=digb)
        return self.msgs.pin(keys=digb, val=msg)


    def cloneTopicIter(self, topic, fn=0):
        """
        Returns iterator of first seen exn messages with attachments for the
        identifier prefix pre starting at first seen order number, fn.

        """
        if hasattr(topic, 'encode'):
            topic = topic.encode("utf-8")

        for (key, dig) in self.getIoSetItemsIter(self.tpcs, key=topic, ion=fn):
            topic, ion = dbing.unsuffix(key)
            if msg := self.msgs.get(keys=dig):
                yield ion, topic, msg.encode("utf-8")



class MailboxServer(doing.DoDoer):
    """
    Message storage for Witnesses.  Provides an inbox service for storing messages for an identifier.

    """

    def __init__(self, mbx: Mailboxer, app=None, **kwa):
        """
        Create Mailbox server for storing messages on a Witness for a witnessed
        identifier.

        Parameters:
             app(falcon.App): REST app to register routes with

        """

        self.mbx = mbx
        self.rxbs = bytearray()

        self.app = app if app is not None else falcon.App(cors_enable=True)

        self.app.add_route("/req/mbx", self)

        doers = []

        super(MailboxServer, self).__init__(doers=doers, **kwa)

    def on_get(self, req, rep):
        """
        Handles GET requests as a stream of SSE events

        Parameters:
              req (Request) Falcon HTTP request
              rep (Response) Falcon HTTP response

        """
        pre = req.params["i"]
        pt = req.params["topics"]

        topics = dict()
        for t in pt:
            key, val = t.split("=")
            topics[key] = int(val)

        rep.stream = self.mailboxGenerator(pre=pre, topics=topics, resp=rep)

    def on_post(self, req, rep):
        """
        Handles GET requests as a stream of SSE events

        Parameters:
              req (Request) Falcon HTTP request
              rep (Response) Falcon HTTP response

        """
        cr = httping.parseCesrHttpRequest(req=req)

        query = cr.payload['q']
        topics = query['topics']
        pre = query["pre"]

        rep.stream = self.mailboxGenerator(pre=pre, topics=topics, resp=rep)


    @helping.attributize
    def mailboxGenerator(self, me, pre=None, topics=None, resp=None):
        """

        Parameters:
            me:
            pre:
            topics:
            resp:

        """
        me._status = http.httping.OK

        headers = Hict()
        headers['Content-Type'] = "text/event-stream"
        headers['Cache-Control'] = "no-cache"
        headers['Connection'] = "keep-alive"
        me._headers = headers

        yield b'retry: 1000\n'
        while True:
            for topic, idx in topics.items():
                key = pre + topic
                for fn, _, msg in self.mbx.cloneTopicIter(key, idx):
                    data = bytearray("id: {}\nevent: {}\ndata: ".format(fn, topic).encode("utf-8"))
                    data.extend(msg)
                    data.extend(b'\n\n')
                    idx = idx + 1
                    yield data
                topics[topic] = idx

            yield b''


class Respondant(doing.DoDoer):
    """
    Respondant processes buffer of response messages from inbound 'exn' messages and
    routes them to the appropriate mailbox.  If destination has witnesses, send response to
    one of the (randomly selected) witnesses.  Otherwise store the response in the recipients
    mailbox locally.

    """

    def __init__(self, hab, reps=None, cues=None, mbx=None, **kwa):
        """
        Creates Respondant that uses local environment to find the destination KEL and stores
        peer to peer messages in mbx, the mailboxer

        Parameters:
            hab (Habitat):  local environment
            mbx (Mailboxer): storage for local messages

        """
        self.reps = reps if reps is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.hab = hab
        self.mbx = mbx if mbx is not None else Mailboxer(name=hab.name)

        doers = [doing.doify(self.responseDo), doing.doify(self.cueDo)]
        super(Respondant, self).__init__(doers=doers, **kwa)


    def responseDo(self, tymth=None, tock=0.0, **opts):
        """
        Doifiable Doist compatibile generator method to process response messages from `exn` handlers.
        If dest is not in local environment, ignore the response (for now).  If dest has witnesses,
        pick one at random and send the response to that witness for storage in the recipients mailbox
        on that witness.  Otherwise this is a peer to peer HTTP message and should be stored in a mailbox
        locally for the recipient.

        Usage:
            add result of doify on this method to doers list
        """

        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while self.reps:
                rep = self.reps.popleft()
                recipient = rep["dest"]
                exn = rep["rep"]
                topic = rep["topic"]

                print("sending to", recipient)
                print(exn.pretty())
                while recipient not in self.hab.kevers:
                    yield self.tock

                kever = self.hab.kevers[recipient]
                if kever is None:
                    logger.Error("unable to reply, dest {} not found".format(recipient))
                    continue

                if len(kever.wits) == 0:
                    msg = bytearray(exn.raw)
                    msg.extend(self.hab.sanction(exn))
                    self.mbx.storeMsg(topic=recipient, msg=msg)
                else:
                    wit = random.choice(kever.wits)
                    loc = obtaining.getwitnessbyprefix(wit)

                    client = http.clienting.Client(hostname=loc.ip4, port=loc.http)
                    clientDoer = http.clienting.ClientDoer(client=client)

                    self.extend([clientDoer])

                    fwd = forwarding.forward(pre=recipient, serder=exn, topic=topic)
                    msg = bytearray(fwd.raw)
                    msg.extend(self.hab.sanction(exn))

                    httping.createCESRRequest(msg, client, date=exn.ked["dt"])

                    while not client.responses:
                        yield self.tock

                    self.remove([clientDoer])

                yield  # throttle just do one cue at a time
            yield



    def cueDo(self, tymth=None, tock=0.0, **opts):
        """
         Returns doifiable Doist compatibile generator method (doer dog) to process
            Kevery and Tevery cues deque

        Usage:
            add result of doify on this method to doers list
        """
        yield  # enter context
        while True:
            while self.cues:  # iteratively process each cue in cues
                msg = bytearray()
                cue = self.cues.popleft()
                cueKin = cue["kin"]  # type or kind of cue

                if cueKin in ("receipt",):  # cue to receipt a received event from other pre
                    serder = cue["serder"]  # Serder of received event for other pre
                    msg.extend(self.hab.receipt(serder))
                    self.mbx.storeMsg(topic=serder.preb+b'/receipt', msg=msg)
                elif cueKin in ("replay",):
                    dest = cue["dest"]
                    msgs = cue["msgs"]
                    self.mbx.storeMsg(topic=dest+'/replay', msg=msgs)

                yield self.tock

            yield self.tock



# -*- encoding: utf-8 -*-
"""
KERI
keri.app.agenting module

"""
import json
import random

import falcon
from falcon import media
from hio.base import doing
from hio.core import http
from hio.core.tcp import clienting
from hio.help import decking
from orderedset import OrderedSet as oset

from . import httping, grouping
from .. import help
from .. import kering
from ..app import obtaining
from ..core import eventing, parsing, scheming, coring
from ..db import dbing
from ..help import helping
from ..help.helping import nowIso8601
from ..peer import exchanging
from ..vc import proving, handling
from ..vdr import registering, issuing, verifying

logger = help.ogler.getLogger()


class WitnessReceiptor(doing.DoDoer):
    """
    Sends messages to all current witnesses of given identifier (from hab) and waits
    for receipts from each of those witnesses and propagates those receipts to each
    of the other witnesses after receiving the complete set.

    Removes all Doers and exits as Done once all witnesses have been sent the entire
    receipt set.  Could be enhanced to have a `once` method that runs once and cleans up
    and an `all` method that runs and waits for more messages to receipt.

    """

    def __init__(self, hab, msg=None, klas=None, **kwa):
        """
        For the current event, gather the current set of witnesses, send the event,
        gather all receipts and send them to all other witnesses

        Parameters:
            hab: Habitat of the identifier to populate witnesses
            msg: is the message to send to all witnesses.
                 Defaults to sending the latest KEL event if msg is None

        """
        self.hab = hab
        self.msg = msg
        self.klas = klas if klas is not None else HttpWitnesser
        super(WitnessReceiptor, self).__init__(doers=[doing.doify(self.receiptDo)], **kwa)

    def receiptDo(self, tymth=None, tock=0.0, **opts):
        """
        Returns doifiable Doist compatible generator method (doer dog)

        Usage:
            add result of doify on this method to doers list

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters

        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)
        sn = self.hab.kever.sn
        wits = self.hab.kever.wits

        if len(wits) == 0:
            return True

        msg = self.msg if self.msg is not None else self.hab.makeOwnEvent(sn=sn)
        ser = coring.Serder(raw=msg)

        witers = []
        for wit in wits:
            witer = self.klas(hab=self.hab, wit=wit)
            witers.append(witer)
            witer.msgs.append(bytearray(msg))  # make a copy
            self.extend([witer])

            _ = (yield self.tock)

        dgkey = dbing.dgKey(ser.preb, ser.digb)
        while True:
            wigs = self.hab.db.getWigs(dgkey)
            if len(wigs) == len(wits):
                break
            _ = yield self.tock

        # generate all rct msgs to send to all witnesses
        wigers = [coring.Siger(qb64b=bytes(wig)) for wig in wigs]
        rserder = eventing.receipt(pre=ser.pre,
                                   sn=sn,
                                   dig=ser.dig)
        rctMsg = eventing.messagize(serder=rserder, wigers=wigers)

        # this is a little brute forcey and can be improved by gathering receipts
        # along the way and passing them out as we go and only sending the
        # required ones here
        for witer in witers:
            witer.msgs.append(bytearray(rctMsg))
            _ = (yield self.tock)

        total = len(witers) * 2
        count = 0
        while count < total:
            for witer in witers:
                count += len(witer.sent)
            _ = (yield self.tock)

        self.remove(witers)
        return True


class WitnessInquisitor(doing.DoDoer):
    """
    Sends messages to all current witnesses of given identifier (from hab) and waits
    for receipts from each of those witnesses and propagates those receipts to each
    of the other witnesses after receiving the complete set.

    Removes all Doers and exits as Done once all witnesses have been sent the entire
    receipt set.  Could be enhanced to have a `once` method that runs once and cleans up
    and an `all` method that runs and waits for more messages to receipt.

    """

    def __init__(self, hab, msgs=None, wits=None, klas=None, **kwa):
        """
        For all msgs, select a random witness from Habitat's current set of witnesses
        send the msg and process all responses (KEL replays, RCTs, etc)

        Parameters:
            hab: Habitat of the identifier to use to identify witnesses
            msgs: is the message buffer to process and send to one random witness.

        """
        self.hab = hab
        self.wits = wits
        self.klas = klas if klas is not None else HttpWitnesser
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.smsgs = oset()

        super(WitnessInquisitor, self).__init__(doers=[doing.doify(self.receiptDo), doing.doify(self.msgDo)], **kwa)

    def receiptDo(self, tymth=None, tock=1.0, **opts):
        """
        Returns doifiable Doist compatible generator method (doer dog)

        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        wits = self.wits if self.wits is not None else self.hab.kever.wits
        if len(wits) == 0:
            raise kering.ConfigurationError("Must be used with an identifier that has witnesses")

        witers = []
        for wit in wits:
            witer = self.klas(hab=self.hab, wit=wit, lax=True, local=False)
            witers.append(witer)

        self.extend(witers)

        while True:
            while not self.smsgs:
                yield self.tock

            msg = self.smsgs.pop()
            witer = random.choice(witers)
            witer.msgs.append(bytearray(msg))

            yield

    def msgDo(self, tymth=None, tock=0.0, **opts):
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while not self.msgs:
                yield self.tock

            msg = self.msgs.popleft()
            self.smsgs.add(msg)


    def query(self, pre, r="logs", sn=0):
        msg = self.hab.query(pre, res=r, query=dict())  # Query for remote pre Event
        self.msgs.append(bytes(msg))

    def telquery(self, ri, i, r="tels"):
        msg = self.hab.query(i, res=r, query=dict(ri=ri))  # Query for remote pre Event
        self.msgs.append(bytes(msg))


class WitnessPublisher(doing.DoDoer):
    """
    Sends messages to all current witnesses of given identifier (from hab) and exits.

    Removes all Doers and exits as Done once all witnesses have been sent the message.
    Could be enhanced to have a `once` method that runs once and cleans up
    and an `all` method that runs and waits for more messages to receipt.

    """

    def __init__(self, hab, msg, wits=None, klas=None, **kwa):
        """
        For the current event, gather the current set of witnesses, send the event,
        gather all receipts and send them to all other witnesses

        Parameters:
            hab: Habitat of the identifier to populate witnesses
            msg: is the message to send to all witnesses.
                 Defaults to sending the latest KEL event if msg is None

        """
        self.hab = hab
        self.msg = msg
        self.wits = wits if wits is not None else self.hab.kever.wits
        self.klas = klas if klas is not None else HttpWitnesser
        super(WitnessPublisher, self).__init__(doers=[doing.doify(self.sendDo)], **kwa)

    def sendDo(self, tymth=None, tock=0.0, **opts):
        """
        Returns doifiable Doist compatible generator method (doer dog)

        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if len(self.wits) == 0:
            return True

        witers = []
        for wit in self.wits:
            witer = self.klas(hab=self.hab, wit=wit)
            witers.append(witer)
            witer.msgs.append(bytearray(self.msg))  # make a copy so everyone munges their own
            self.extend([witer])

            _ = (yield self.tock)

        total = len(witers)
        count = 0
        while count < total:
            for witer in witers:
                count += len(witer.sent)
            _ = (yield self.tock)

        self.remove(witers)
        return True


class TCPWitnesser(doing.DoDoer):
    """

    """

    def __init__(self, hab, wit, msgs=None, sent=None, doers=None, **kwa):
        """
        For the current event, gather the current set of witnesses, send the event,
        gather all receipts and send them to all other witnesses

        Parameters:
            hab: Habitat of the identifier to populate witnesses

        """
        self.hab = hab
        self.wit = wit
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.sent = sent if sent is not None else decking.Deck()
        self.parser = None
        doers = doers if doers is not None else []
        doers.extend([doing.doify(self.receiptDo)])

        self.kevery = eventing.Kevery(db=self.hab.db,
                                      **kwa)

        super(TCPWitnesser, self).__init__(doers=doers)

    def receiptDo(self, tymth=None, tock=0.0):
        """
        Returns doifiable Doist compatible generator method (doer dog)

        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        loc = obtaining.getwitnessbyprefix(self.wit)
        client = clienting.Client(host=loc.ip4, port=loc.tcp)
        self.parser = parsing.Parser(ims=client.rxbs,
                                     framed=True,
                                     kvy=self.kevery)

        clientDoer = clienting.ClientDoer(client=client)
        self.extend([clientDoer, doing.doify(self.msgDo)])

        while True:
            while not self.msgs:
                yield self.tock

            msg = self.msgs.popleft()

            client.tx(msg)  # send to connected remote

            while client.txbs:
                yield self.tock

            self.sent.append(msg)
            yield self.tock

    def msgDo(self, tymth=None, tock=0.0, **opts):
        """
        Returns doifiable Doist compatible generator method (doer dog) to process
            incoming message stream of .kevery

        Doist Injected Attributes:
            g.tock = tock  # default tock attributes
            g.done = None  # default done state
            g.opts

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters


        Usage:
            add result of doify on this method to doers list
        """
        yield from self.parser.parsator()  # process messages continuously


class HttpWitnesser(doing.DoDoer):
    """
    Interacts with Witnesses on HTTP and SSE for sending events and receiving receipts

    """

    def __init__(self, hab, wit, msgs=None, sent=None, doers=None, **kwa):
        """
        For the current event, gather the current set of witnesses, send the event,
        gather all receipts and send them to all other witnesses

        Parameters:
            hab: Habitat of the identifier to populate witnesses

        """
        self.hab = hab
        self.wit = wit
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.sent = sent if sent is not None else decking.Deck()
        self.parser = None
        doers = doers if doers is not None else []
        doers.extend([doing.doify(self.msgDo), doing.doify(self.responseDo)])

        self.kevery = eventing.Kevery(db=self.hab.db,
                                      lax=False,
                                      local=True)

        loc = obtaining.getwitnessbyprefix(self.wit)

        self.client = http.clienting.Client(hostname=loc.ip4, port=loc.http)
        clientDoer = http.clienting.ClientDoer(client=self.client)

        doers.extend([clientDoer])

        super(HttpWitnesser, self).__init__(doers=doers, **kwa)

    def msgDo(self, tymth=None, tock=0.0):
        """
        Returns doifiable Doist compatible generator method (doer dog)

        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while not self.msgs:
                yield self.tock

            msg = self.msgs.popleft()

            httping.createCESRRequest(msg, self.client)
            while self.client.requests:
                yield self.tock

            yield self.tock

    def responseDo(self, tymth=None, tock=0.0):
        """
        Processes responses from client and adds them to sent cue

        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while not self.client.responses:
                rep = self.client.respond()
                self.sent.append(rep)
                yield
            yield


class KiwiServer(doing.DoDoer):
    """
    Routes for handling UI requests for Credential issuance/revocation and presentation requests

    """

    def __init__(self, hab, controller, rep, verifier, gdoer, issuers=None, issuerCues=None, cues=None, app=None,
                 insecure=False,
                 **kwa):
        """
        Create a KIWI web server for Agents capable of performing KERI and ACDC functions for the controller
        of an identifier.

        Parameters:
            hab Habitat is the environment of the identifier prefix
            controller qb64 is the identifier prefix that can send commands to this web server:
            rep Respondant that routes responses to the appropriate mailboxes
            verifier is Verifier that process credentials
            gdoe is decking.Deck of msgs to send to a MultisigDoer
            issuers is dict of credential Issuers keyed by regk of credential Registry
            wallet is Wallet for local storage of credentials
            cues is Deck from Kevery handling key events:
            app falcon.App to register handlers with:
            insecure bool is True to allow requests without verifying KERI Http Signature Header,
                defaults to False

        """
        self.hab = hab
        self.controller = controller
        self.rep = rep
        self.verifier = verifier if verifier is not None else verifying.Verifier(hab=self.hab, name=hab.name)
        self.app = app if app is not None else falcon.App(cors_enable=True)
        self.issuers = issuers if issuers is not None else dict()
        self.typ = scheming.JSONSchema()

        if insecure:
            self.app.add_middleware(httping.InsecureSignatureComponent())
        else:
            self.app.add_middleware(httping.SignatureValidationComponent(hab=hab, pre=controller))

        self.app.req_options.media_handlers.update(media.Handlers())
        self.app.resp_options.media_handlers.update(media.Handlers())
        self.cues = cues if cues is not None else decking.Deck()
        self.issuerCues = issuerCues if issuerCues is not None else decking.Deck()

        self.app.add_route("/id", self, suffix="id")
        self.app.add_route("/registry/incept", self, suffix="registry_incept")
        self.registryIcpr = registering.RegistryInceptDoer(hab=hab)
        self.app.add_route("/credential/apply", self, suffix="apply")
        self.app.add_route("/credential/issue", self, suffix="issue")
        self.app.add_route("/credential/revoke", self, suffix="revoke")

        self.app.add_route("/credentials/issued", self, suffix="credentials_issued")
        self.app.add_route("/credentials/received", self, suffix="credentials_received")

        self.app.add_route("/presentation/request", self, suffix="request")

        self.app.add_route("/multisig/incept", self, suffix="multisig_incept")
        self.app.add_route("/multisig/rotate", self, suffix="multisig_rotate")
        self.app.add_route("/multisig", self, suffix="multisig")
        self.gdoer = gdoer

        self.app.add_route("/delegate/incept", self, suffix="delegate_incept")
        self.app.add_route("/delegate/rotate", self, suffix="delegate_rotate")
        # self.delcptr = delegating.InceptDoer(name="")
        # self.delrotr = delegating.RotateDoer(hab=hab)

        self.witq = WitnessInquisitor(hab=hab, klas=HttpWitnesser)

        doers = [self.witq, self.registryIcpr, doing.doify(self.verifierDo), doing.doify(
            self.issuerDo), doing.doify(self.escrowDo)]

        super(KiwiServer, self).__init__(doers=doers, **kwa)


    def verifierDo(self, tymth, tock=0.0, **opts):
        """
        Process cues from Verifier coroutine

            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters
        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.verifier.cues:
                cue = self.verifier.cues.popleft()
                cueKin = cue["kin"]

                if cueKin == "saved":
                    creder = cue["creder"]
                    proof = cue["proof"]

                    logger.info("Credential: %s, Schema: %s,  Saved", creder.said, creder.schema)
                    logger.info(creder.pretty())
                    print("Credential: {}, Schema: {},  Saved".format(creder.said, creder.schema))
                    print(creder.pretty())

                    recpt = creder.subject["i"]

                    craw = bytearray(creder.raw)
                    if len(proof) % 4:
                        raise ValueError("Invalid attachments size={}, nonintegral"
                                         " quadlets.".format(len(proof)))
                    craw.extend(coring.Counter(code=coring.CtrDex.AttachedMaterialQuadlets,
                                               count=(len(proof) // 4)).qb64b)

                    craw.extend(proof)

                    pl = dict(
                        vc=[handling.envelope(craw)]
                    )

                    exn = exchanging.exchange(route="/credential/issue", payload=pl)
                    self.rep.reps.append(dict(dest=recpt, rep=exn, topic="credential"))


                elif cueKin == "query":
                    qargs = cue["q"]
                    self.witq.query(**qargs)

                elif cueKin == "telquery":
                    qargs = cue["q"]
                    self.witq.telquery(**qargs)

                yield self.tock
            yield self.tock

    def escrowDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method

        Usage:
            add result of doify on this method to doers list

        Processes the Groupy escrow for group icp, rot and ixn request messages.

        """
        # start enter context
        yield  # enter context
        while True:
            issuers = dict(self.issuers)
            for _, issuer in issuers.items():
                issuer.processEscrows()
                yield self.tock

            self.verifier.processEscrows()
            yield self.tock


    def issuerDo(self, tymth, tock=0.0, **opts):
        """
        Process cues from credential issue coroutine

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters
        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.issuerCues:
                cue = self.issuerCues.popleft()
                cueKin = cue['kin']
                if cueKin == "send":
                    tevt = cue["msg"]
                    witSender = WitnessPublisher(hab=self.hab, msg=tevt)
                    self.extend([witSender])

                    while not witSender.done:
                        _ = yield self.tock

                    self.remove([witSender])
                elif cueKin == "kevt":
                    kevt = cue["msg"]
                    witDoer = WitnessReceiptor(hab=self.hab, msg=kevt)
                    self.extend([witDoer])

                    while not witDoer.done:
                        yield self.tock

                    self.remove([witDoer])
                elif cueKin == "multisig":
                    msg = dict(
                        op=cue["op"],
                        data=cue["data"],
                        reason=cue["reason"]
                    )
                    self.gdoer.append(msg)
                elif cueKin == "logEvent":
                    print("TEL event saved")


                yield self.tock
            yield self.tock


    def on_post_apply(self, req, rep):
        """
        Apply for a credential with the given credential fields.

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        Body:
            name: The human readable name of the new registry to create

        """
        body = json.loads(req.context.raw)
        schema = body.get("schema")
        typ = body.get("type")
        issuer = body.get("issuer")
        values = body.get("values")

        apply = handling.credential_apply(issuer=issuer, schema=schema, typ=typ, formats=[], body=values)

        exn = exchanging.exchange(route="/credential/apply", payload=apply)
        self.rep.reps.append(dict(dest=issuer, rep=exn, topic="credential"))


    def on_post_issue(self, req, rep):
        """
        Initiate a credential issuanace from this agent's identifier

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        Body:
            name: The human readable name of the new registry to create

        """
        body = json.loads(req.context.raw)
        registry = body.get("registry")
        schema = body.get("schema")
        source = body.get("source")
        recipientIdentifier = body.get("recipient")
        notify = body["notify"] if "notify" in body else True

        issuer = self.getIssuer(name=registry)
        if issuer is None:
            rep.status = falcon.HTTP_400
            rep.text = "Credential registry {} does not exist.  It must be created before issuing " \
                       "credentials".format(registry)
            return

        types = ["VerifiableCredential", body.get("type")]

        data = body.get("credentialData")
        dt = data["dt"] if "dt" in data else nowIso8601()

        d = dict(
            d="",
            i=recipientIdentifier,
            dt=dt,
            t=types,
        )

        d |= data

        group = self.hab.group()
        if group is None:
            pre = self.hab.pre
        else:
            pre = group.gid

        creder = proving.credential(issuer=pre,
                                    schema=schema,
                                    subject=d,
                                    source=source,
                                    status=issuer.regk)
        try:
            issuer.issue(creder=creder, dt=dt)
        except kering.MissingAnchorError:
            logger.info("Missing anchor from credential issuance due to multisig identifier")

        craw = self.hab.endorse(creder)
        proving.parseCredential(ims=craw, verifier=self.verifier)

        if notify and group:
            for aid in group.aids:
                if aid != self.hab.pre:
                    if aid not in self.hab.kevers:
                        self.witq.query(aid)
                    msg = dict(
                        schema=schema,
                        source=source,
                        recipient=recipientIdentifier,
                        typ=body.get("type"),
                        data=d,
                    )
                    exn = exchanging.exchange(route="/multisig/issue", payload=msg)
                    self.rep.reps.append(dict(dest=aid, rep=exn, topic="multisig"))

        rep.status = falcon.HTTP_200
        rep.data = creder.pretty().encode("utf-8")

    def on_post_revoke(self, req, rep):
        """

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        Body:
            name: The human readable name of the new registry to create

        """
        body = json.loads(req.context.raw)
        registry = body.get("registry")
        said = body.get("said")

        issuer = self.getIssuer(name=registry)
        if issuer is None:
            rep.status = falcon.HTTP_400
            rep.text = "Credential registry {} does not exist.  It must be created before issuing " \
                       "credentials".format(registry)
            return

        try:
            issuer.revoke(vcdig=said)
        except kering.ValidationError as ex:
            rep.status = falcon.HTTP_CONFLICT
            rep.text = ex.args[0]
            return

        rep.status = falcon.HTTP_202

    def on_post_request(self, req, rep):
        """
        HTTP handler for credential presentation request generation

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        Body:
            recipient: SCID of holder of credential to request
            schema:  SAID of schema

        """
        body = json.loads(req.context.raw)
        recipientIdentifier = body.get("recipient")
        schema = body.get("schema")

        pl = dict(
            input_descriptors=[
                dict(x=schema)
            ]
        )

        exn = exchanging.exchange(route="/presentation/request", payload=pl)
        self.rep.reps.append(dict(dest=recipientIdentifier, rep=exn, topic="credential"))

        rep.status = falcon.HTTP_202

    def on_post_multisig_rotate(self, req, rep):
        """

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        Body:
            name: The human readable name of the new registry to create

        """
        body = json.loads(req.context.raw)

        msg = dict(
            sith=None,
            toad=None,
            data=None,
            witnesses=[],
            witness_cuts=[],
            witness_adds=[],
        )

        for key in msg:
            if key in body:
                msg[key] = body[key]

        msg["op"] = grouping.Ops.rot

        self.gdoer.append(msg)

        rep.status = falcon.HTTP_202

    def on_post_multisig_incept(self, req, rep):
        """

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        Body:
            name: The human readable name of the new registry to create

        """
        body = json.loads(req.context.raw)

        if "aids" not in body:
            rep.status = falcon.HTTP_400
            rep.text = "Invalid multisig group inception request, 'aids' is required'"
            return

        aids = body["aids"]

        notify = body["notify"] if "notify" in body else True

        msg = dict(
            aids=aids,
            toad=None,
            witnesses=[],
            isith=None,
            nsith=None)

        for key in msg:
            if key in body:
                msg[key] = body[key]

        if notify:
            for aid in aids:
                if aid != self.hab.pre:
                    if aid not in self.hab.kevers:
                        self.witq.query(aid)
                    exn = exchanging.exchange(route="/multisig/incept", payload=dict(msg), date=helping.nowIso8601())
                    self.rep.reps.append(dict(dest=aid, rep=exn, topic="multisig"))

        msg["op"] = grouping.Ops.icp

        self.gdoer.append(msg)

        rep.status = falcon.HTTP_202

    def on_get_multisig(self, req, rep):
        """
        Return the groups this environment is a part of

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        """
        res = []

        group = self.hab.group()
        if group:
            kever = self.hab.kevers[group.gid]
            ser = kever.serder
            dgkey = dbing.dgKey(ser.preb, ser.digb)
            wigs = self.hab.db.getWigs(dgkey)

            gd = dict(
                prefix=group.gid,
                seq_no=kever.sn,
                aids=group.aids,
                delegated=kever.delegated,
                witnesses=kever.wits,
                public_keys=[verfer.qb64 for verfer in kever.verfers],
                toad=kever.toad,
                isith=kever.tholder.sith,
                receipts=len(wigs)
            )

            res.append(gd)

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(res).encode("utf-8")

    def on_get_credentials_issued(self, req, rep):
        """
        Return the credntials issued by this agent

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        """
        registry = req.params["registry"]
        issuer = self.getIssuer(registry)
        group = self.hab.group()

        if group is None:
            pre = self.hab.pre
        else:
            pre = group.gid

        saids = issuer.reger.issus.get(keys=pre)
        creds = self.get_credentials(issuer, saids)

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(creds).encode("utf-8")


    @staticmethod
    def get_credentials(issuer, saids):
        creds = []
        for saider in saids:
            key = saider.qb64b
            creder = issuer.reger.creds.get(keys=key)

            # TODO:  de-dupe the seals here and extract the signatures
            seals = issuer.reger.seals.get(keys=key)
            prefixer = None
            seqner = None
            diger = None
            sigers = []
            for seal in seals:
                (prefixer, seqner, diger, siger) = seal
                sigers.append(siger)

            status, lastSeen = issuer.tevers[issuer.regk].vcState(key)
            cred = dict(
                sad=creder.crd,
                pre=prefixer.qb64,
                sn=seqner.sn,
                dig=diger.qb64,
                sigers=[sig.qb64 for sig in sigers],
                status=status,
                # lastSeen=lastSeen.dts,
            )

            creds.append(cred)
        return creds


    def on_get_credentials_received(self, req, rep):
        """
        Return the credntials received by this agent

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        """
        registry = req.params["registry"]
        issuer = self.getIssuer(registry)

        group = self.hab.group()

        if group is None:
            pre = self.hab.pre
        else:
            pre = group.gid

        saids = issuer.reger.subjs.get(keys=pre)
        creds = self.get_credentials(issuer, saids)

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(creds).encode("utf-8")

    def on_post_registry_incept(self, req, rep):
        """

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        Body:
            name: The human readable name of the new registry to create

        """
        body = json.loads(req.context.raw)

        if "name" not in body:
            rep.status = falcon.HTTP_400
            rep.text = "name is a required parameter to create a verifiable credential registry"
            return

        msg = dict(name=body["name"])
        self.registryIcpr.msgs.append(msg)

        rep.status = falcon.HTTP_202

    def getIssuer(self, name):
        if name in self.issuers:
            issuer = self.issuers[name]
        else:
            issuer = issuing.Issuer(hab=self.hab, name=name, reger=self.verifier.reger, cues=self.issuerCues)
            self.issuers[name] = issuer

        return issuer

    def on_get_id(self, req, rep):
        """
        Return the groups this environment is a part of

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        """
        res = []

        habs = self.hab.db.habs.getItemIter()
        for (name,), habr in habs:
            kever = self.hab.kevers[habr.prefix]
            ser = kever.serder
            dgkey = dbing.dgKey(ser.preb, ser.digb)
            wigs = self.hab.db.getWigs(dgkey)

            gd = dict(
                name=name,
                prefix=habr.prefix,
                seq_no=kever.sn,
                delegated=kever.delegated,
                witnesses=kever.wits,
                public_keys=[verfer.qb64 for verfer in kever.verfers],
                toad=kever.toad,
                isith=kever.tholder.sith,
                receipts=len(wigs)
            )

            res.append(gd)

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(res).encode("utf-8")

    def on_post_delegate_incept(self, req, rep):
        rep.status = falcon.HTTP_202

    def on_post_delegate_rotate(self, req, rep):
        rep.status = falcon.HTTP_202



from keri.app import keeping, habbing
from keri.db import basing


def openHabitat(name="test", **kwa):
    ks = keeping.Keeper(name=name, temp=False)  # not opened by default, doer opens
    ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
    db = basing.Baser(name=name, temp=False, reload=True)  # not opened by default, doer opens
    dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

    # setup habitat
    hab = habbing.Habitat(name=name, ks=ks, db=db, temp=False, create=False, **kwa)
    habDoer = habbing.HabitatDoer(habitat=hab)  # setup doer

    doers = [ksDoer, dbDoer, habDoer]

    return hab, doers



import argparse

from keri.app import directing
from keri.app.cli.commands.vc.registry import incept

parser = argparse.ArgumentParser(description="Create a multisig registry for a group identifier")
parser.set_defaults(handler=lambda args: demo(args))


def demo(args):

    ms1 = incept.RegistryInceptor(name="multisig1", registryName="gleif", estOnly=False, noBackers=True, baks=[])
    ms2 = incept.RegistryInceptor(name="multisig2", registryName="gleif", estOnly=False, noBackers=True, baks=[])
    ms3 = incept.RegistryInceptor(name="multisig3", registryName="gleif", estOnly=False, noBackers=True, baks=[])

    doers = [ms1, ms2, ms3]
    directing.runController(doers, expire=0.0)



# -*- encoding: utf-8 -*-
"""
KERI
keri.app.forwarding module

module for enveloping and forwarding KERI message
"""

import random

from hio.base import doing
from hio.help import decking

from keri.app import agenting
from keri.core import coring, eventing


class Postman(doing.DoDoer):
    """
    DoDoer that wraps any KERI event (KEL, TEL, Peer to Peer) in a `fwd` envelope and
    delivers to sends them to one of the target recipient's witnesses for store and forward
    to the intended recipient

    """

    def __init__(self, hab, evts=None, klas=None, **kwa):
        self.hab = hab
        self.evts = evts if evts is not None else decking.Deck()
        self.klas = klas if klas is not None else agenting.HttpWitnesser

        doers = [doing.doify(self.deliverDo)]
        super(Postman, self).__init__(doers=doers, **kwa)

    def deliverDo(self, tymth=None, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method that processes
                   a queue of messages and envelopes them in a `fwd` message
                   and sends them to one of the witnesses of the recipient for
                   store and forward.

        Usage:
            add result of doify on this method to doers list
        """

        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while self.evts:
                evt = self.evts.popleft()
                recp = evt["recipient"]
                tpc = evt["topic"]
                srdr = evt["serder"]
                act = evt["attachment"]

                # TODO: sign the forward message with a SAID signature and
                #  combine it with the provided attachments on the envelope
                #  so the envelope can get verified and opened and this
                #  message can be extracted and stored.
                fwd = forward(pre=recp, topic=tpc, serder=srdr)
                ims = bytearray(fwd.raw)
                ims.extend(act)

                kever = self.hab.kevers[recp]
                wit = random.choice(kever.wits)

                witer = self.klas(hab=self.hab, wit=wit)
                witer.msgs.append(bytearray(ims))  # make a copy
                self.extend([witer])

                while not witer.sent:
                    _ = (yield self.tock)

                yield self.tock

            yield self.tock

    def send(self, recipient, topic, msg):
        """
        Utility function to queue a msg on the Postman's buffer for
        enveloping and forwarding to a witness

        Parameters:
            recipient is identifier prefix qb64 of the intended recipient
            msg is bytes of signed KERI event message to envelope and forward:

        """
        serder = coring.Serder(raw=msg)
        del msg[:serder.size]

        self.evts.append(dict(recipient=recipient, topic=topic, serder=serder, attachment=bytearray(msg)))


def forward(pre, serder, topic=None, version=coring.Version, kind=coring.Serials.json):
    """
    Returns serder of forward event message.
    Utility function to automate creation of forward events.

     Parameters:
        pre is identifier prefix qb64 of recipient of message
        serder is Serder of message to wrap in forward envelope
        version is Version instance
        kind is serialization kind

    """

    vs = coring.Versify(version=version, kind=kind, size=0)
    ilk = eventing.Ilks.fwd

    r = pre + "/" + topic if topic is not None else pre

    ked = dict(v=vs,
               t=ilk,
               r=r,
               a=serder.ked,
               )


    return eventing.Serder(ked=ked)  # return serialized ked




# -*- encoding: utf-8 -*-
"""
keri.vc.proving module

"""

import json
import logging
from collections.abc import Iterable
from typing import Union

import cbor2 as cbor
import msgpack

from .. import help, kering
from ..core import coring
from ..core.coring import (Serials, sniff, Versify, Deversify, Rever, Counter,
                           CtrDex, Prefixer, Seqner, Diger, Siger, Saider, Ids)
from ..core.parsing import Parser, Colds
from ..db import subing
from ..kering import Version, VersionError, ShortageError, DeserializationError, ColdStartError, ExtractionError

KERI_REGISTRY_TYPE = "KERICredentialRegistry"

logger = help.ogler.getLogger()


def credential(schema,
               issuer,
               subject,
               status=None,
               source=None,
               version=Version,
               kind=Serials.json):
    """
    Returns Credentialer of new credential
        schema is SAID of schema for this credential
        issuer is the identifier prefix of the issuer
        subject is dict of the values being assigned to the subject of this credential
        source is list of source credentials to which this credential is chained
        version is Version instance
        kind is serialization kind

    """
    vs = Versify(version=version, kind=kind, size=0)

    source = source if source is not None else []

    vc = dict(
        v=vs,
        d="",
        s=schema,
        i=issuer,
        a={},
        p=source
    )

    if status is not None:
        subject["ri"] = status

    _, sad = coring.Saider.saidify(sad=subject, kind=kind, label=coring.Ids.d)
    vc["a"] = sad

    return Credentialer(crd=vc)


def parseCredential(ims, verifier):
    parsator = allParsator(ims=ims, verifier=verifier)

    while True:
        try:
            next(parsator)
        except StopIteration:
            break


def allParsator(ims, verifier):
    if not isinstance(ims, bytearray):
        ims = bytearray(ims)  # so make bytearray copy

    while ims:  # only process until ims empty
        try:
            done = yield from credParsator(ims=ims,
                                           verifier=verifier,
                                           )

        except kering.SizedGroupError as ex:  # error inside sized group
            # processOneIter already flushed group so do not flush stream
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception("Parser msg extraction error: %s\n", ex.args[0])
            else:
                logger.error("Parser msg extraction error: %s\n", ex.args[0])

        except (kering.ColdStartError, kering.ExtractionError) as ex:  # some extraction error
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception("Parser msg extraction error: %s\n", ex.args[0])
            else:
                logger.error("Parser msg extraction error: %s\n", ex.args[0])
            del ims[:]  # delete rest of stream to force cold restart

        except (kering.ValidationError, Exception) as ex:  # non Extraction Error
            # Non extraction errors happen after successfully extracted from stream
            # so we don't flush rest of stream just resume
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception("Parser msg non-extraction error: %s\n", ex)
            else:
                logger.error("Parser msg non-extraction error: %s\n", ex)
        yield

    return True



def credentialParsator(ims, verifier):
    if not isinstance(ims, bytearray):
        ims = bytearray(ims)  # so make bytearray copy

        while True:  # continuous stream processing never stop
            try:
                done = yield from credParsator(ims=ims,
                                               verifier=verifier)

            except kering.SizedGroupError as ex:  # error inside sized group
                # processOneIter already flushed group so do not flush stream
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Parser msg extraction error: %s\n", ex.args[0])
                else:
                    logger.error("Parser msg extraction error: %s\n", ex.args[0])

            except (kering.ColdStartError, kering.ExtractionError) as ex:  # some extraction error
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Parser msg extraction error: %s\n", ex.args[0])
                else:
                    logger.error("Parser msg extraction error: %s\n", ex.args[0])
                del ims[:]  # delete rest of stream to force cold restart

            except (kering.ValidationError, Exception) as ex:  # non Extraction Error
                # Non extraction errors happen after successfully extracted from stream
                # so we don't flush rest of stream just resume
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Parser msg non-extraction error: %s\n", ex.args[0])
                else:
                    logger.error("Parser msg non-extraction error: %s\n", ex.args[0])
            yield



def credParsator(ims=b'', verifier=None):
    """
    Parse the ims bytearray as a CESR Proof Format verifiable credential

    Parameters:
        ims (bytearray) of serialized incoming verifiable credential in CESR Proof Format.
        verifier (verifying.Verifier) verifier and storage for the verified credential
        typ (JSONSchema) class for resolving schema references:

    """

    while not ims:
        yield

    cold = Parser.sniff(ims)  # check for spurious counters at front of stream
    if cold in (Colds.txt, Colds.bny):  # not message error out to flush stream
        # replace with pipelining here once CESR message format supported.
        raise kering.ColdStartError("Expecting message counter tritet={}"
                                    "".format(cold))

    while True:  # extract and deserialize message from ims
        try:
            creder = Credentialer(raw=ims)
        except ShortageError as e:
            raise e
        else:
            del ims[:creder.size]
            break


    # extract attachments must start with counter so know if txt or bny.
    while not ims:
        yield
    cold = Parser.sniff(ims)
    if cold is Colds.msg:
        raise ColdStartError("unable to parse VC, attachments expected")

    ctr = Parser.extract(ims=ims, klas=Counter, cold=cold)
    if ctr.code != CtrDex.AttachedMaterialQuadlets:
        raise ExtractionError("Invalid attachment to VC {}, expected {}"
                              "".format(ctr.code, CtrDex.AttachedMaterialQuadlets))

    pags = ctr.count * 4
    if len(ims) != pags:
        raise ShortageError("VC proof attachment invalid length {}, expected {}"
                            "".format(len(ims), pags))

    prefixer, seqner, diger, isigers = parseProof(ims=ims)

    try:
        verifier.processCredential(creder, prefixer, seqner, diger, isigers)
    except AttributeError as ex:
        raise kering.ValidationError("No verifier to process so dropped credential"
                                     "= {}.".format(creder.pretty()))

    return True  # done state


def parseProof(ims=b''):
    cold = Parser.sniff(ims)
    if cold is Colds.msg:
        raise ColdStartError("unable to parse VC, attachments expected")

    ctr = Parser.extract(ims=ims, klas=Counter, cold=cold)
    if ctr.code != CtrDex.TransIdxSigGroups or ctr.count != 1:
        raise ExtractionError("Invalid attachment to VC {}, expected one {}"
                              "".format(ctr.code, CtrDex.TransIdxSigGroups))

    prefixer = Parser.extract(ims=ims, klas=Prefixer)
    seqner = Parser.extract(ims=ims, klas=Seqner)
    diger = Parser.extract(ims=ims, klas=Diger)

    ictr = Parser.extract(ims=ims, klas=Counter)
    if ictr.code != CtrDex.ControllerIdxSigs:
        raise ExtractionError("Invalid attachment to VC {}, expected {}"
                              "".format(ctr.code, CtrDex.ControllerIdxSigs))

    isigers = []
    for i in range(ictr.count):
        isiger = Parser.extract(ims=ims, klas=Siger)
        isigers.append(isiger)

    return prefixer, seqner, diger, isigers


def buildProof(prefixer, seqner, diger, sigers):
    """

    Parameters:
        prefixer (Prefixer) Identifier of the issuer of the credential
        seqner (Seqner) is the sequence number of the event used to sign the credential
        diger (Diger) is the digest of the event used to sign the credential
        sigers (list) are the cryptographic signatures on the credential

    """

    prf = bytearray()
    prf.extend(Counter(CtrDex.TransIdxSigGroups, count=1).qb64b)
    prf.extend(prefixer.qb64b)
    prf.extend(seqner.qb64b)
    prf.extend(diger.qb64b)

    prf.extend(Counter(code=CtrDex.ControllerIdxSigs, count=len(sigers)).qb64b)
    for siger in sigers:
        prf.extend(siger.qb64b)

    return prf


class Credentialer:
    """
    Credentialer is for creating a W3C Verifiable Credential embedded in a CESR Proof Format
    proof

    """

    def __init__(self, raw=b'', crd=None, kind=None, code=coring.MtrDex.Blake3_256):
        """
        Creates a serializer/deserializer for a Verifiable Credential in CESR Proof Format

        requires either raw or (crd and kind) to load credential from serialized form or in memory

        Parameters:
            raw (bytes) is raw credential
            crd (dict) is populated credential
            typ is schema type
            version is Version instance
            kind is serialization kind

        """
        self._code = code  # need default code for .diger

        if raw:  # deserialize raw using property setter
            self.raw = raw  # raw property setter does the deserialization
        elif crd:  # serialize ked using property setter
            self._kind = kind
            self.crd = crd  # ked property setter does the serialization
        else:
            raise ValueError("Improper initialization need raw or ked.")


    @staticmethod
    def _inhale(raw):
        """
        Parse raw according to serialization type and return dict of values, kind, version and size

        """
        kind, version, size = sniff(raw)
        if version != Version:
            raise VersionError("Unsupported version = {}.{}, expected {}."
                               "".format(version.major, version.minor, Version))
        if len(raw) < size:
            raise ShortageError("Need more bytes.")

        if kind == Serials.json:
            try:
                crd = json.loads(raw[:size].decode("utf-8"))
            except Exception:
                raise DeserializationError("Error deserializing JSON: {}"
                                           "".format(raw[:size].decode("utf-8")))

        elif kind == Serials.mgpk:
            try:
                crd = msgpack.loads(raw[:size])
            except Exception:
                raise DeserializationError("Error deserializing MGPK: {}"
                                           "".format(raw[:size]))

        elif kind == Serials.cbor:
            try:
                crd = cbor.loads(raw[:size])
            except Exception:
                raise DeserializationError("Error deserializing CBOR: {}"
                                           "".format(raw[:size]))

        else:
            raise DeserializationError("Error deserializing unsupported kind: {}"
                                       "".format(raw[:size].decode("utf-8")))

        return crd, kind, version, size

    @staticmethod
    def _exhale(crd, kind=None):
        """
        Create serialized format from dict of VC values.  Returns raw, kind, dict of values and version

        """

        knd, version, size = Deversify(crd["v"])  # extract kind and version
        if version != Version:
            raise ValueError("Unsupported version = {}.{}".format(version.major,
                                                                  version.minor))

        crd["d"] = Saider.Dummy * coring.Matter.Codes[coring.MtrDex.Blake3_256].fs

        if not kind:
            kind = knd

        if kind not in Serials:
            raise ValueError("Invalid serialization kind = {}".format(kind))

        raw = coring.dumps(crd, kind)
        size = len(raw)

        match = Rever.search(raw)
        if not match or match.start() > 12:
            raise ValueError("Invalid version string in raw = {}".format(raw))

        fore, back = match.span()

        # update vs with latest kind version size
        vs = Versify(version=version, kind=kind, size=size)
        # replace old version string in raw with new one
        raw = b'%b%b%b' % (raw[:fore], vs.encode("utf-8"), raw[back:])
        if size != len(raw):  # substitution messed up
            raise ValueError("Malformed version string size = {}".format(vs))
        crd["v"] = vs

        saider = Saider(sad=crd, code=coring.MtrDex.Blake3_256, label=Ids.d)
        crd["d"] = saider.qb64

        raw = coring.dumps(crd, kind)

        return raw, kind, crd, version, saider

    @property
    def kind(self):
        """ kind property getter"""
        return self._kind

    @property
    def raw(self):
        """ raw gettter bytes of serialized type """
        return self._raw

    @raw.setter
    def raw(self, raw):
        """ raw property setter """
        crd, kind, version, size = self._inhale(raw=raw)
        self._raw = bytes(raw[:size])  # crypto ops require bytes not bytearray
        self._crd = crd
        self._kind = kind
        self._version = version
        self._size = size
        self._saider = Saider(qb64=self._crd[Ids.d], code=coring.MtrDex.Blake3_256, label=Ids.d)

    @property
    def crd(self):
        """ crd dict property getter"""
        return self._crd

    @crd.setter
    def crd(self, crd):
        """ ked property setter  assumes ._kind """
        raw, kind, crd, version, saider = self._exhale(crd=crd, kind=self._kind)
        size = len(raw)
        self._raw = raw[:size]
        self._crd = crd
        self._kind = kind
        self._size = size
        self._version = version
        self._saider = saider

    @property
    def size(self):
        """ size property getter"""
        return self._size

    @property
    def saider(self):
        """ saider property getter"""
        return self._saider

    @property
    def said(self):
        """ said property getter, relies on saider """
        return self.saider.qb64

    @property
    def issuer(self):
        """ issuer property getter"""
        return self.crd["i"]

    @property
    def schema(self):
        """ schema property getter"""
        return self.crd["s"]

    @property
    def subject(self):
        """ subject property getter"""
        return self.crd["a"]

    @property
    def status(self):
        """ status property getter"""
        return self.crd["a"]["ri"]

    def pretty(self):
        """
        Returns str JSON of .ked with pretty formatting
        """
        return json.dumps(self.crd, indent=1)


class CrederSuber(subing.Suber):
    """
    Sub class of Suber where data is serialized Credentialer instance
    Automatically serializes and deserializes using Credentialer methods

    """

    def __init__(self, *pa, **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
        """
        super(CrederSuber, self).__init__(*pa, **kwa)


    def put(self, keys: Union[str, Iterable], val: Credentialer):
        """
        Puts val at key made from keys. Does not overwrite

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (Credentialer): instance

        Returns:
            result (bool): True If successful, False otherwise, such as key
                              already in database.
        """
        return (self.db.putVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=val.raw))


    def pin(self, keys: Union[str, Iterable], val: Credentialer):
        """
        Pins (sets) val at key made from keys. Overwrites.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (Credentialer): instance

        Returns:
            result (bool): True If successful. False otherwise.
        """
        return (self.db.setVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=val.raw))


    def get(self, keys: Union[str, Iterable]):
        """
        Gets Serder at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            Credentialer:
            None: if no entry at keys

        Usage:
            Use walrus operator to catch and raise missing entry
            if (creder := mydb.get(keys)) is None:
                raise ExceptionHere
            use creder here

        """
        val = self.db.getVal(db=self.sdb, key=self._tokey(keys))
        return Credentialer(raw=bytes(val)) if val is not None else None


    def rem(self, keys: Union[str, Iterable]):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise
        """
        return self.db.delVal(db=self.sdb, key=self._tokey(keys))


    def getItemIter(self, keys: Union[str, Iterable]=b""):
        """
        Return iterator over the all the items in subdb

        Returns:
            iterator: of tuples of keys tuple and val coring.Serder for
            each entry in db

        """
        for key, val in self.db.getTopItemIter(db=self.sdb, key=self._tokey(keys)):
            yield self._tokeys(key), Credentialer(raw=bytes(val))



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.witness module

Witness command line interface
"""
import argparse
import logging

from hio.base import doing
from keri import __version__, kering
from keri import help
from keri.app import indirecting, storing
from keri.app.cli.common import existing
from keri.core import scheming
from keri.peer import exchanging
from keri.vc import walleting, handling
from keri.vdr import verifying

d = "Runs KERI Agent controller.\n"
d += "Example:\nagent -t 5621\n"
parser = argparse.ArgumentParser(description=d)
parser.set_defaults(handler=lambda args: launch(args))
parser.add_argument('-V', '--version',
                    action='version',
                    version=__version__,
                    help="Prints out version of script runner.")
parser.add_argument('-n', '--name',
                    action='store',
                    default="wallet",
                    help="Name of controller. Default is wallet.")



def launch(args):
    help.ogler.level = logging.INFO
    help.ogler.reopen(name=args.name, temp=True, clear=True)

    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Wallet for %s."
                ".******\n\n", args.name)

    runWallet(name=args.name)

    logger.info("\n******* Ended Wallet for %s."
                ".******\n\n", args.name)


def runWallet(name="wallet"):
    """
    Setup and run one wallet
    """

    hab, doers = existing.openHabitat(name=name)
    verifier = verifying.Verifier(hab=hab, name=name)
    wallet = walleting.Wallet(db=verifier.reger, name=name)
    walletDoer = walleting.WalletDoer(hab=hab, verifier=verifier)

    jsonSchema = scheming.JSONSchema(resolver=scheming.jsonSchemaCache)
    issueHandler = handling.IssueHandler(hab=hab, verifier=verifier, typ=jsonSchema)
    requestHandler = handling.RequestHandler(wallet=wallet, typ=jsonSchema)
    exchanger = exchanging.Exchanger(hab=hab, handlers=[issueHandler, requestHandler])

    mbx = storing.Mailboxer(name=name)
    rep = storing.Respondant(hab=hab, mbx=mbx)
    mdir = indirecting.MailboxDirector(hab=hab, exc=exchanger, rep=rep, topics=["/receipt", "/replay", "/credential"])

    doers.extend([exchanger, mdir, rep, walletDoer])

    try:
        tock = 0.03125
        doist = doing.Doist(limit=0.0, tock=tock, real=True)
        doist.do(doers=doers)
    except kering.ConfigurationError:
        print(f"prefix for {name} does not exist, incept must be run first", )



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from hio import help
from keri.app import habbing
from keri.app.cli.common import displaying
from keri.core import coring
from keri.kering import ConfigurationError

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a prefix')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument("--verbose", "-V", help="print JSON of all current events", action="store_true")


def handler(args):

    name = args.name
    try:
        with habbing.existingHab(name=name) as hab:

            displaying.printIdentifier(hab, hab.pre)

            if len(hab.prefixes) > 1:
                print("\nAdditional Prefixes:")
                for pre in hab.prefixes:
                    if pre == hab.pre:
                        continue
                    print(f"\t{pre}")

            group = hab.group()
            if group:
                print()
                print("Groups:")
                print("\t   {} ({})".format(name, group.gid))
                print()

            if args.verbose:
                cloner = hab.db.clonePreIter(pre=hab.pre, fn=0)  # create iterator at 0
                for msg in cloner:
                    srdr = coring.Serder(raw=msg)
                    print(srdr.pretty())
                    print()

    except ConfigurationError as e:
        print(e)
        print(f"identifier prefix for {name} does not exist, incept must be run first", )
        return -1



import argparse
import logging

from hio.base import doing

from keri import help, kering
from keri.app.cli.commands.agent import start

parser = argparse.ArgumentParser(description="Run a demo collection of multisig agents")
parser.set_defaults(handler=lambda args: demo(args))


def demo(args):
    help.ogler.level = logging.CRITICAL
    help.ogler.reopen(name="keri", temp=True, clear=True)
    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Multisig Agents multisig1, multisig2, multisig3 "
                ".******\n\n")

    # kli agent start --name multisig1 --pre E4Zq5dxbnWKq5K-Bssn4g_qhBbSwNSI2MH4QYnkEUFDM --insecure --tcp 5621
    # --http 5620 --admin-http-port 5623
    doers = start.runAgent(controller="E4Zq5dxbnWKq5K-Bssn4g_qhBbSwNSI2MH4QYnkEUFDM",
                           name="multisig1", insecure=True,
                           tcp=5621,
                           adminHttpPort=5623)

    # kli agent start --name multisig2 --pre E4Zq5dxbnWKq5K-Bssn4g_qhBbSwNSI2MH4QYnkEUFDM --insecure --tcp 5721
    # --http 5720 --admin-http-port 5723
    doers += start.runAgent(controller="E4Zq5dxbnWKq5K-Bssn4g_qhBbSwNSI2MH4QYnkEUFDM",
                            name="multisig2", insecure=True,
                            tcp=5721,
                            adminHttpPort=5723)

    # kli agent start --name multisig3 --pre E4Zq5dxbnWKq5K-Bssn4g_qhBbSwNSI2MH4QYnkEUFDM --insecure --tcp 5821
    # --http 5820 --admin-http-port 5823
    doers += start.runAgent(controller="E4Zq5dxbnWKq5K-Bssn4g_qhBbSwNSI2MH4QYnkEUFDM",
                            name="multisig3", insecure=True,
                            tcp=5821,
                            adminHttpPort=5823)

    try:
        tock = 0.03125
        doist = doing.Doist(limit=0.0, tock=tock, real=True)
        doist.do(doers=doers)
    except kering.ConfigurationError:
        print(f"prefix for {args.name} does not exist, incept must be run first", )

    logger.info("\n******* Ended Multisig Agents multisig1, multisig2, multisig3 "
                ".******\n\n",)



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import json

from hio.base import doing
from keri import kering
from keri.db import basing

from ... import habbing, keeping, agenting, indirecting, directing

parser = argparse.ArgumentParser(description='Rotate keys')
parser.set_defaults(handler=lambda args: interact(args))
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--proto', '-p', help='Protocol to use when propagating ICP to witnesses [tcp|http] (defaults '
                                          'http)', default="tcp")
parser.add_argument('--data', '-d', help='Anchor data, \'@\' allowed', default=[], action="store", required=False)


def interact(args):
    """
    Performs a rotation of the identifier of the environment represented by the provided name parameter

        args (parseargs):  Command line argument

    """
    name = args.name
    if args.data is not None:
        try:
            if args.data.startswith("@"):
                f = open(args.data[1:], "r")
                data = json.load(f)
            else:
                data = json.loads(args.data)
        except json.JSONDecodeError:
            raise kering.ConfigurationError("data supplied must be value JSON to anchor in a seal")

        if not isinstance(data, list):
            data = [data]

    else:
        data = None

    ixnDoer = InteractDoer(name=name, proto=args.proto, data=data)

    doers = [ixnDoer]

    try:
        directing.runController(doers=doers, expire=0.0)
    except kering.ConfigurationError:
        print(f"identifier prefix for {name} does not exist, incept must be run first", )
        return -1



class InteractDoer(doing.DoDoer):
    """
    DoDoer that launches Doers needed to create an interaction event and publication of the event
    to all appropriate witnesses
    """

    def __init__(self, name, proto, data: list = None):
        """
        Returns DoDoer with all registered Doers needed to perform interaction event.

        Parameters:
            name is human readable str of identifier
            proto is tcp or http method for communicating with Witness
            data is list of dicts of committed data such as seals
       """

        self.name = name
        self.proto = proto
        self.data = data

        ks = keeping.Keeper(name=self.name, temp=False)  # not opened by default, doer opens
        self.ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
        db = basing.Baser(name=self.name, temp=False, reload=True)  # not opened by default, doer opens
        self.dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

        self.hab = habbing.Habitat(name=self.name, ks=ks, db=db, temp=False, create=False)
        self.habDoer = habbing.HabitatDoer(habitat=self.hab)  # setup doer
        doers = [self.ksDoer, self.dbDoer, self.habDoer, doing.doify(self.interactDo)]

        super(InteractDoer, self).__init__(doers=doers)


    def interactDo(self, tymth, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)


        msg = self.hab.interact(data=self.data)

        if self.proto == "tcp":
            mbx = None
            witDoer = agenting.WitnessReceiptor(hab=self.hab, klas=agenting.TCPWitnesser, msg=msg)
            self.extend(doers=[witDoer])
            yield self.tock
        else:  # "http"
            mbx = indirecting.MailboxDirector(hab=self.hab)
            witDoer = agenting.WitnessReceiptor(hab=self.hab, klas=agenting.HTTPWitnesser, msg=msg)
            self.extend(doers=[mbx, witDoer])
            yield self.tock

        while not witDoer.done:
            _ = yield self.tock


        print(f'Prefix  {self.hab.pre}')
        print(f'New Sequence No.  {self.hab.kever.sn}')
        for idx, verfer in enumerate(self.hab.kever.verfers):
            print(f'\tPublic key {idx+1}:  {verfer.qb64}')

        toRemove = [self.ksDoer, self.dbDoer, self.habDoer, witDoer]
        if mbx:
            toRemove.append(mbx)

        self.remove(toRemove)

        return



# -*- encoding: utf-8 -*-
"""
KERI
keri.vdr.issuing module

VC issuer support
"""
import json

from hio.base import doing
from hio.help import decking

from keri.vdr import viring
from .. import kering, help
from ..app import grouping, agenting
from ..core import parsing, coring
from ..core.coring import Counter, Seqner, CtrDex, MtrDex, Serder
from ..core.eventing import SealEvent, SealSource, TraitDex
from ..db.dbing import snKey, dgKey
from ..help import helping
from ..vc import proving
from ..vdr import eventing
from ..vdr.viring import Registry, nsKey

logger = help.ogler.getLogger()


class Issuer:
    """
    Issuer provides encapsulation of creating a Verifiable Credential Registry with issuance
    and revocation of VCs against that registry.

    The Registry consists of 1 management TEL for maintaining the state of the registry wrt special
    Backers that can act as witnesses of VC events, and 1 VC TEL for each VC issued that tracks the
    issuance and revocation status of those VCs.

    """

    def __init__(self, hab, name="test", cues=None, reger=None, estOnly=False,
                 temp=False, **kwa):
        """
        Initialize Instance

        Parameters:
            name is the alias for this issuer
            hab is Habitat instance of local controller's context
            reger is Registry database instance for controller's credentials
            tevers is a dict of Tever instances keys by qb64 prefix of registry
            noBackers is boolean True to allow specification of TEL specific backers
            backers is the initial list of backer prefixes qb64 for VCs in the Registry
            toad is int or str hex of witness threshold
            estOnly is boolean True for forcing rotation events for every TEL event.
        """

        self.hab = hab
        self.name = name
        self.estOnly = estOnly
        self.cues = cues if cues is not None else decking.Deck()
        self.regk = None

        self.reger = reger if reger is not None else Registry(name=self.hab.name, temp=temp)
        self.inited = False

        # save init kwy word arg parameters as ._inits in order to later finish
        # init setup elseqhere after databases are opened if not below
        self._inits = kwa

        if self.hab.inited:
            self.setup(**self._inits)

    def setup(self, *, noBackers=False, baks=None, toad=None, ):

        ex = self.reger.regs.get(keys=self.name)
        if ex is not None:
            self.regk = ex.registryKey

        if self.regk is None:
            self.regi = 0

            self.noBackers = noBackers

            # save backers locally for now.  will be managed by tever when implemented
            self.backers = baks if baks is not None else []

            self.cnfg = [TraitDex.NoBackers] if self.noBackers else []

            group = self.hab.group()
            if group is None:
                pre = self.hab.pre
            else:
                pre = group.gid

            self.regser = eventing.incept(pre,
                                          baks=self.backers,
                                          toad=toad,
                                          cnfg=self.cnfg,
                                          code=MtrDex.Blake3_256)
            self.regk = self.regser.pre
            self.reger.regs.put(keys=self.name,
                                val=viring.RegistryRecord(registryKey=self.regk))

            self.tvy = eventing.Tevery(reger=self.reger, db=self.hab.db, regk=self.regk, local=True)
            self.psr = parsing.Parser(framed=True, kvy=self.hab.kvy, tvy=self.tvy)

            try:
                self.anchorMsg(self.regser)
            except kering.MissingAnchorError:
                logger.info("Credential registry missing anchor for inception = {}".format(self.regser.ked))
        else:
            self.tvy = eventing.Tevery(reger=self.reger, db=self.hab.db, regk=self.regk, local=True)
            self.psr = parsing.Parser(framed=True, kvy=self.hab.kvy, tvy=self.tvy)

            clone = self.reger.clonePreIter(self.regk)
            for msg in clone:
                self.psr.parseOne(ims=msg)

            if self.regk not in self.tevers:
                raise kering.ConfigurationError("Improper Issuer inception for "
                                                "pre={}.".format(self.regk))

            tever = self.tevers[self.regk]
            self.noBackers = tever.noBackers
            self.backers = tever.baks
            self.regi = int(tever.serder.ked["s"], 16)

        self.inited = True

    @property
    def tevers(self):
        """
        Returns .db.tevers
        """
        return self.reger.tevers

    def rotate(self, toad=None, cuts=None, adds=None):
        """
        Rotate backer list for registry

        Parameters:
            toad is int or str hex of backer threshold after cuts and adds
            cuts is list of qb64 pre of backers to be removed from witness list
            adds is list of qb64 pre of backers to be added to witness list
        """

        if self.noBackers:
            raise ValueError("Attempt to rotate registry {} that does not support backers".format(self.regk))

        serder = eventing.rotate(dig=self.regser.dig, regk=self.regk, sn=self.regi + 1, toad=toad, baks=self.backers,
                                 adds=adds, cuts=cuts)

        self.regser = serder

        self.anchorMsg(serder)

        tever = self.tevers[self.regk]
        self.backers = tever.baks
        self.regi = int(tever.serder.ked["s"], 16)

        return True

    def issue(self, creder, dt=None):
        """
        Create and process an iss or bis message event

        Parameters:
            creder is hash digest of vc content qb64

        """
        vcdig = creder.said
        craw = self.hab.endorse(creder)

        if self.noBackers:
            serder = eventing.issue(vcdig=vcdig, regk=self.regk, dt=dt)
        else:
            serder = eventing.backerIssue(vcdig=vcdig, regk=self.regk, regsn=self.regi, regd=self.regser.diger.qb64,
                                          dt=dt)

        self.anchorMsg(serder, reason=craw.decode("utf-8"))

        return True

    def revoke(self, vcdig, dt=None):
        """

        Create and process iss message event

        Parameters:
            vcdig is hash digest of vc content qb64

        """

        vckey = nsKey([self.regk, vcdig])
        vcser = self.reger.getTel(snKey(pre=vckey, sn=0))
        if vcser is None:
            raise kering.ValidationError("Invalid revoke of {} that has not been issued "
                                         "pre={}.".format(vcdig, self.regk))
        ievt = self.reger.getTvt(dgKey(pre=vckey, dig=vcser))
        iserder = Serder(raw=bytes(ievt))

        if self.noBackers:
            serder = eventing.revoke(vcdig=vcdig, regk=self.regk, dig=iserder.dig, dt=dt)
        else:
            serder = eventing.backerRevoke(vcdig=vcdig, regk=self.regk, regsn=self.regi, regd=self.regser.diger.qb64,
                                           dig=iserder.dig, dt=dt)

        self.anchorMsg(serder)

        return True

    @staticmethod
    def messagize(serder, seal):
        msg = bytearray(serder.raw)
        msg.extend(Counter(CtrDex.SealSourceCouples, count=1).qb64b)
        msg.extend(Seqner(sn=seal.s).qb64b)
        msg.extend(seal.d.encode("utf-8"))

        return msg

    def anchorMsg(self, serder, reason=None, seal=None):

        group = self.hab.group()

        rseal = SealEvent(serder.pre, serder.ked["s"], serder.dig)
        rseal = rseal._asdict()

        if group is None:
            if self.estOnly:
                kevt = self.hab.rotate(data=[rseal])
            else:
                kevt = self.hab.interact(data=[rseal])

            seal = SealSource(s=self.hab.kever.sn, d=self.hab.kever.serder.dig)
            tevt = self.messagize(serder=serder, seal=seal)

            self.psr.parseOne(ims=bytearray(tevt))  # make copy as kvr deletes
            self.cues.extend([
                dict(
                    kin="kevt",
                    msg=kevt
                ),
                dict(
                    kin="send",
                    msg=tevt
                ),
            ])

        else:
            if seal is None:
                op = grouping.Ops.rot if self.estOnly else grouping.Ops.ixn
                mmsg = dict(kin="multisig", op=op, data=[rseal], reason=reason)
                self.cues.append(mmsg)

                self.escrow(serder)
                raise kering.MissingAnchorError("anchor not provided for multisig")
            else:
                tevt = self.messagize(serder=serder, seal=seal)
                self.psr.parseOne(ims=bytearray(tevt))  # make copy as kvr deletes

                self.cues.append(dict(kin="logEvent", msg=tevt))

    def escrow(self, serder):
        """
        Save Issuer event for future process when anchor become available
        Parameters:
           serder: (Serder) is event to escrow

        """
        self.reger.mase.add(self.regk, serder.raw)

    def processEscrows(self):
        """
        Process

        """
        for (regk,), raw in self.reger.mase.getItemIter():
            serder = coring.Serder(raw=raw.encode("utf-8"))

            if "ri" in serder.ked:
                tev = self.tevers[serder.ked["ri"]]
                pre = tev.pre
            else:
                pre = serder.ked["ii"]

            sn = serder.sn
            dig = serder.dig

            seal = None
            for evts in self.hab.db.clonePreIter(pre=pre):
                eserder = coring.Serder(raw=evts)
                if "a" in eserder.ked:
                    ancs = eserder.ked["a"]
                    if len(ancs) != 1:
                        continue

                    anc = ancs[0]
                    spre = anc["i"]
                    ssn = int(anc["s"])
                    sdig = anc["d"]

                    if spre == serder.ked["i"] and ssn == sn \
                            and dig == sdig:
                        seal = SealSource(s=eserder.sn, d=eserder.dig)
                        break

            if not seal:
                continue

            try:
                self.anchorMsg(serder, seal=seal, reason=None)
            except kering.MissingAnchorError as ex:
                logger.exception("Issuer unescrow failed event from escrow = {}", ex.args[0])
            except Exception as ex:
                logger.exception("Issuer unescrow failed event from escrow = {}", ex.args[0])
                self.reger.mase.rem(regk, raw)
            else:  # unescrow succeeded, remove from escrow
                # We don't remove all escrows at pre,sn because some might be
                # duplicitous so we process remaining escrows in spite of found
                # valid event escrow.
                self.reger.mase.rem(regk, raw)
                logger.info("Issuer unescrow succeeded in valid event: "
                            "event=\n%s\n", json.dumps(serder.ked, indent=1))


class IssuerDoer(doing.DoDoer):
    """
    Basic Issuer Doer  to initialize inception events of the registry

    Inherited Attributes:
        .done is Boolean completion state:
            True means completed
            Otherwise incomplete. Incompletion maybe due to close or abort.

    Attributes:
        .issuer is Issuer subclass

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is float, desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Methods:
        .wind  injects ._tymth dependency from associated Tymist to get its .tyme
        .__call__ makes instance callable
            Appears as generator function that returns generator
        .do is generator method that returns generator
        .enter is enter context action method
        .recur is recur context action method or generator method
        .exit is exit context method
        .close is close context method
        .abort is abort context method

    Hidden:
        ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
        ._tock is hidden attribute for .tock property
    """

    def __init__(self, hab, issuer, verifier, msgs=None, cues=None, **kwa):
        """
        Parameters:
           issuer (Issuer): instance
        """
        self.hab = hab
        self.issuer = issuer
        self.verifier = verifier
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.gdoer = grouping.MultiSigGroupDoer(hab=hab)

        doers = [self.gdoer, doing.doify(self.issueDo), doing.doify(self.issuerDo), doing.doify(self.escrowDo)]

        super(IssuerDoer, self).__init__(doers=doers, **kwa)


    def enter(self, **kwargs):
        if not self.issuer.inited:
            self.issuer.setup(**self.issuer._inits)
        super(IssuerDoer, self).enter(**kwargs)


    def issueDo(self, tymth, tock=0.0, **kwa):
        """
        Returns:  doifiable Doist compatible generator method for creating a registry
        and sending its inception and anchoring events to witnesses or backers

        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)  # finish enter context

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                schema = msg["schema"]
                typ = msg["typ"]
                source = msg["source"]
                recipient = msg["recipient"]
                data = msg["data"]

                dt = data["dt"] if "dt" in data else None

                types = ["VerifiableCredential", typ]

                d = dict(
                    i="",
                    type=types,
                    si=recipient,
                    dt=helping.nowIso8601()
                )

                d |= data

                group = self.hab.group()
                if group is None:
                    pre = self.hab.pre
                else:
                    name, group = group
                    pre = group.gid

                creder = proving.credential(issuer=pre,
                                            schema=schema,
                                            subject=d,
                                            source=source,
                                            status=self.issuer.regk)



                try:
                    self.issuer.issue(creder=creder, dt=dt)
                except kering.MissingAnchorError:
                    logger.info("Missing anchor from credential issuance due to multisig identifier")

                craw = self.hab.endorse(creder)
                proving.parseCredential(ims=craw, verifier=self.verifier)

                yield self.tock

            yield self.tock

    def issuerDo(self, tymth, tock=0.0, **opts):
        """
        Process cues from credential issue coroutine

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters
        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.issuer.cues:
                cue = self.issuer.cues.popleft()

                cueKin = cue['kin']
                if cueKin == "send":
                    tevt = cue["msg"]
                    witSender = agenting.WitnessPublisher(hab=self.hab, msg=tevt)
                    self.extend([witSender])

                    while not witSender.done:
                        _ = yield self.tock

                    self.remove([witSender])
                elif cueKin == "kevt":
                    kevt = cue["msg"]
                    witDoer = agenting.WitnessReceiptor(hab=self.hab, msg=kevt)
                    self.extend([witDoer])

                    while not witDoer.done:
                        yield self.tock

                    self.remove([witDoer])

                    self.cues.append(dict(kin="finished"))
                elif cueKin == "multisig":
                    msg = dict(
                        op=cue["op"],
                        group=cue["group"],
                        data=cue["data"],
                        reason=cue["reason"]
                    )
                    self.gdoer.msgs.append(msg)
                elif cueKin == "logEvent":
                    print("TEL event saved")


                yield self.tock

            yield self.tock

    def escrowDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method

        Usage:
            add result of doify on this method to doers list

        Processes the Groupy escrow for group icp, rot and ixn request messages.

        """
        # start enter context
        yield  # enter context
        while True:
            self.issuer.processEscrows()
            self.verifier.processEscrows()
            yield




# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands.multisig module

"""

import argparse
import json
import logging
import sys
from json import JSONDecodeError

from hio.base import doing
from hio.help import decking

from keri import help
from keri.app import directing, indirecting, grouping, agenting
from keri.app.cli.common import displaying, existing

# help.ogler.level = logging.INFO
# help.ogler.reopen(name="hio", temp=True, clear=True)

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a group identifier prefix')
parser.set_defaults(handler=lambda args: inceptMultisig(args))
parser.add_argument('--name', '-n', help='Human readable environment reference for local identifier', required=True)
parser.add_argument('--group', '-g', help="Human readable environment reference for group identifier", required=True)
parser.add_argument('--file', '-f', help='Filename to use to create the identifier', default="", required=True)
parser.add_argument('--proto', '-p', help='Protocol to use when propagating ICP to witnesses [tcp|http] (defaults '
                                          'http)', default="http")


def inceptMultisig(args):
    """
    Reads the config file into a MultiSigInceptOptions dataclass and creates and signs the inception
    event for the group identifier.  If signatures are provided in the options file, the event is submitted
    to its witnesses and receipts are collected.

    Parameters:
        args: Parsed arguments from the command line

    """

    # help.ogler.level = logging.INFO
    # help.ogler.reopen(name=args.name, temp=True, clear=True)

    try:
        f = open(args.file)
        opts = json.load(f)
    except FileNotFoundError:
        print("config file", args.file, "not found")
        sys.exit(-1)
    except JSONDecodeError:
        print("config file", args.file, "not valid JSON")
        sys.exit(-1)

    name = args.name
    group = args.group

    icpDoer = GroupMultisigIncept(name=name, group=group, proto=args.proto, **opts)

    doers = [icpDoer]
    directing.runController(doers=doers, expire=0.0)


class GroupMultisigIncept(doing.DoDoer):

    def __init__(self, name, group, **kwa):
        self.name = name
        self.hab, doers = existing.openHabitat(name=name)
        self.msg = kwa
        self.msg["group"] = group

        topics = ['/receipt', '/multisig', '/replay']
        if "delpre" in self.msg:
            topics.append('/delegate')


        self.mbd = indirecting.MailboxDirector(hab=self.hab, topics=topics)
        self.icpr = grouping.MultiSigGroupDoer(hab=self.hab, ims=self.mbd.ims)
        self.witq = agenting.WitnessInquisitor(hab=self.hab, klas=agenting.HttpWitnesser)


        doers.extend([self.witq, doing.doify(self.cueDo), self.icpr, self.mbd])
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.inceptDo)])

        super(GroupMultisigIncept, self).__init__(doers=doers)

    def inceptDo(self, tymth, tock=0.0, **opts):
        # enter context
        yield self.tock

        self.msg["op"] = grouping.Ops.icp
        self.icpr.msgs.append(self.msg)

        while not self.icpr.cues:
            yield self.tock

        rep = self.icpr.cues.popleft()


        print()
        print("Group Identifier Inception Complete:")
        displaying.printIdentifier(self.hab, rep["pre"])

        self.remove(self.toRemove)

    def cueDo(self, tymth, tock=0.0, **opts):
        """

        Handle cues coming out of our external Mailbox listener and forward to controller
        mailbox if appropriate

        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.mbd.cues:
                cue = self.mbd.cues.popleft()
                cueKin = cue["kin"]  # type or kind of cue
                if cueKin in ("psUnescrow",):
                    if self.icpr.msgToSend is not None:

                        witRctDoer = agenting.WitnessReceiptor(hab=self.hab, msg=self.icpr.msgToSend,
                                                               klas=agenting.TCPWitnesser)
                        self.extend([witRctDoer])

                        while not witRctDoer.done:
                            _ = yield self.tock

                        self.remove([witRctDoer])

                        serder = cue["serder"]
                        self.icpr.cues.append(dict(pre=serder.pre))
                elif cue["kin"] == "delegatage":
                    delpre = cue["delpre"]
                    self.witq.query(delpre)

                yield self.tock
            yield self.tock




# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands.delegate module

"""
import argparse
import json
import sys
from dataclasses import dataclass
from json import JSONDecodeError

from hio.base import doing

from keri import help
from keri.app import directing, delegating

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a seal for creating a delegated identifier')
parser.set_defaults(handler=lambda args: incept(args))
parser.add_argument('--name', '-n', help='Human readable environment reference', required=True)
parser.add_argument('--file', '-f', help='Filename to use to create the identifier', default="", required=True)
parser.add_argument('--seal', '-s', help='Filename to write seal', default="", required=False)


@dataclass
class DelegateOptions:
    """
    Options dataclass loaded from the file parameter to this command line function.
    Represents all the options needed to create a delegated identifier

    """
    delpre: str
    salt: str
    transferable: bool
    wits: list
    icount: int
    isith: str
    ncount: int
    nsith: str


def incept(args):
    """
    Reads the config file into a CreateDelegateOptions dataclass and creates
    delegated identifier prefixes and events

    Parameters:
        args: Parsed arguments from the command line

    """
    try:
        f = open(args.file)
        config = json.load(f)
        opts = DelegateOptions(**config)

    except FileNotFoundError:
        print("config file", args.file, "not found")
        sys.exit(-1)
    except JSONDecodeError:
        print("config file", args.file, "not valid JSON")
        sys.exit(-1)

    name = args.name
    sealFile = args.seal

    kwa = opts.__dict__
    icpDoer = DelegateInceptDoer(name=name, sealFile=sealFile, **kwa)

    doers = [icpDoer]
    directing.runController(doers=doers, expire=0.0)


class DelegateInceptDoer(doing.DoDoer):
    """
    DoDoer instance that launches the environment and dependencies needed to create and disseminate
    the inception event for a delegated identifier.
    """

    def __init__(self, name, sealFile, **kwa):
        """
        Creates the DoDoer needed to create the seal for a delegated identifier.

        Parameters
            name (str): Name of the local identifier environment

        """

        self.name = name
        self.msg = kwa

        self.dcptr = delegating.InceptDoer(name=name)
        doers = [doing.doify(self.inceptDo), self.dcptr]
        self.toRemove = list(doers)
        super(DelegateInceptDoer, self).__init__(doers=doers, **kwa)

    def inceptDo(self, tymth, tock=0.0, **kwa):
        yield self.tock
        self.dcptr.msgs.append(self.msg)

        while not self.dcptr.cues:
            yield self.tock

        rep = self.dcptr.cues.popleft()
        print(f'Successfully delegated:')
        print(f'\tPrefix: {rep["pre"]}')
        print(f'\tDelegator: {rep["delegator"]}')

        self.remove(self.toRemove)

        return
